"""This module has all of the magic incantations we use to get a pseudo-CST
from the cleancopy tree-sitter binding.
"""
from __future__ import annotations

import logging
from collections.abc import Callable
from collections.abc import Iterator
from decimal import Decimal
from enum import StrEnum
from typing import Literal
from typing import cast

import tree_sitter_cleancopy
from tree_sitter import Language as TSLanguage
from tree_sitter import Node as TSNode
from tree_sitter import Parser as TSParser
from tree_sitter import Tree as TSTree

from cleancopy._cst.nodes import CSTAnnotation
from cleancopy._cst.nodes import CSTDocument
from cleancopy._cst.nodes import CSTDocumentNode
from cleancopy._cst.nodes import CSTDocumentNodeContentEmbedding
from cleancopy._cst.nodes import CSTDocumentNodeContentRichtext
from cleancopy._cst.nodes import CSTDocumentNodeMetadata
from cleancopy._cst.nodes import CSTDocumentNodeTitle
from cleancopy._cst.nodes import CSTEmptyLine
from cleancopy._cst.nodes import CSTFmtBracketLink
from cleancopy._cst.nodes import CSTFmtBracketMetadata
from cleancopy._cst.nodes import CSTFmtBracketMetadataContainer
from cleancopy._cst.nodes import CSTFormattingMarker
from cleancopy._cst.nodes import CSTLineBreak
from cleancopy._cst.nodes import CSTList
from cleancopy._cst.nodes import CSTListItem
from cleancopy._cst.nodes import CSTListItemOLIndexContainer
from cleancopy._cst.nodes import CSTMetadataAssignment
from cleancopy._cst.nodes import CSTMetadataBool
from cleancopy._cst.nodes import CSTMetadataDecimal
from cleancopy._cst.nodes import CSTMetadataInt
from cleancopy._cst.nodes import CSTMetadataMention
from cleancopy._cst.nodes import CSTMetadataMissing
from cleancopy._cst.nodes import CSTMetadataNull
from cleancopy._cst.nodes import CSTMetadataReference
from cleancopy._cst.nodes import CSTMetadataStr
from cleancopy._cst.nodes import CSTMetadataTag
from cleancopy._cst.nodes import CSTMetadataTarget
from cleancopy._cst.nodes import CSTMetadataValue
from cleancopy._cst.nodes import CSTMetadataVariable
from cleancopy._cst.nodes import CSTNode
from cleancopy._cst.nodes import CSTRichtext
from cleancopy.spectypes import InlineFormatting
from cleancopy.spectypes import ListType

CLEANCOPY_LANGUAGE = TSLanguage(tree_sitter_cleancopy.language())
CLEANCOPY_PARSER = TSParser(CLEANCOPY_LANGUAGE)

logger = logging.getLogger(__name__)
_TS_NODE_REGISTRY = {}
_AUTOCLOSE_GRAMMAR_TYPE = 'fmt_autoclose'
_EMPTY_LINE_GRAMMAR_TYPE = 'empty_line'


class _TSNamedField(StrEnum):
    DOCUMENT = 'document'
    ROOT_NODE_CONTENT = 'root_node_content'

    # Note: also includes fmt_autoclose
    RICHTEXT_LINE = 'richtext_line'
    RICHTEXT_FMT_BRACKET = 'fmt_bracket'
    RICHTEXT_FMT_TARGET = 'target'

    NODE = 'node'
    NODE_TITLE = 'title'
    NODE_METADATA = 'metadata'
    NODE_CONTENT = 'node_content'
    EMBEDDING_CONTENT = 'embedding_content'

    INLINE_METADATA = 'inline_metadata'
    METADATA_ASSIGNMENT = 'declaration'
    METADATA_KEY = 'key'
    METADATA_VALUE = 'value'

    LIST_OL = 'list_ol'
    LIST_UNOL = 'list_unol'
    LIST_ITEM = 'list_item'
    LIST_OL_INDEX = 'index'
    # Frustrating artifact of the treesitter grammar
    ANNOTATION_LINE = 'annotation'
    ANNOTATION_LINE_CONTENT = 'annotation_content'

    EMPTY_LINE = 'empty_line'


class _TSValueTypes(StrEnum):
    STRING1 = 'STRING1'
    STRING2 = 'STRING2'
    NULL = 'NULL'
    MISSING = 'MISSING'
    TRUE = 'TRUE'
    FALSE = 'FALSE'
    INT = 'integer'
    DECIMAL = 'decimal'
    MENTION = 'mention'
    TAG = 'tag'
    VARIABLE = 'variable'
    REF = 'ref'
    URI = 'uri'


class _TSRichtextTypes(StrEnum):
    PLAINTEXT = 'plaintext'
    FMT_PRE = 'ext_fmt_pre'
    FMT_UNDERLINE = 'ext_fmt_underline'
    FMT_STRONG = 'ext_fmt_strong'
    FMT_EMPHASIS = 'ext_fmt_emphasis'
    FMT_STRIKE = 'ext_fmt_strike'
    FMT_BRACKET_LINK_ANON = 'fmt_bracket_anon_link'
    FMT_BRACKET_LINK = 'fmt_bracket_named_link'
    FMT_BRACKET_METADATA = 'fmt_bracket_metadata'


def _get_string_quote_style(
        string_type:
            Literal[_TSValueTypes.STRING1] | Literal[_TSValueTypes.STRING2]
        ) -> Literal[1] | Literal[2]:
    if string_type == _TSValueTypes.STRING1:
        return 1
    else:
        return 2


def _register[T, **P](
        field_name: _TSNamedField
        ) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """Use this as a value dispatch for tree sitter node types."""
    if field_name in _TS_NODE_REGISTRY:
        raise ValueError('Already registered!', field_name)

    def decorator(func: Callable[P, T]) -> Callable[P, T]:
        _TS_NODE_REGISTRY[field_name] = func
        return func

    return decorator


def handle_node(field_name: _TSNamedField, node: TSNode) -> CSTNode | None:
    handler = _TS_NODE_REGISTRY.get(field_name)

    if handler is None:
        print(_TS_NODE_REGISTRY)
        print(field_name)
        logger.warning(
            'Node type %s not yet implemented: grammar_name %s, field_name %s',
            node.type, node.grammar_name, field_name)
        return None

    return handler(node)


def parse(document: bytes) -> CSTDocument:
    ts_tree = CLEANCOPY_PARSER.parse(document)
    return to_cst(ts_tree)


def to_cst(tree: TSTree) -> CSTDocument:
    """Converts a CST parse tree into a Document."""
    root_node = tree.root_node

    grammar_type = root_node.type
    if grammar_type != _TSNamedField.DOCUMENT:
        raise ValueError('AST tree must have document as outermost node!')

    maybe_document = handle_node(_TSNamedField.DOCUMENT, root_node)
    if maybe_document is None:
        raise RuntimeError('Failed to process document!')
    elif not isinstance(maybe_document, CSTDocument):
        raise TypeError('Parse tree had no document at root!', maybe_document)
    return maybe_document


@_register(_TSNamedField.DOCUMENT)
def process_document_root(node: TSNode) -> CSTDocument:
    raw_root_node_body, *invalid_extras = _process_named_children(node)

    if invalid_extras:
        raise TypeError('Document root had invalid extras!', invalid_extras)
    if not isinstance(raw_root_node_body, CSTDocumentNodeContentRichtext):
        raise TypeError('Invalid child of document root!', raw_root_node_body)

    # Note: as per spec, we skip any leading emptylines
    first_nonempty_encountered = False
    filtered_content: list[
        CSTEmptyLine | CSTList | CSTAnnotation | CSTRichtext | CSTDocumentNode
    ] = []
    for child_cst_node in raw_root_node_body.content:
        if isinstance(child_cst_node, CSTEmptyLine):
            if first_nonempty_encountered:
                filtered_content.append(child_cst_node)
        else:
            filtered_content.append(child_cst_node)

    # Note that the AST document might have a title and/or metadata, but ONLY
    # the AST -- that doesn't get added in the CST!
    root_node = CSTDocumentNode(
        title=[],
        metadata=[],
        content=CSTDocumentNodeContentRichtext(content=filtered_content))
    return CSTDocument(root=root_node)


@_register(_TSNamedField.ROOT_NODE_CONTENT)
@_register(_TSNamedField.NODE_CONTENT)
def process_node_richtext_content(
        node: TSNode) -> CSTDocumentNodeContentRichtext:
    result = CSTDocumentNodeContentRichtext(content=[])

    for child_cst_node in _process_named_children(node):
        _add_or_merge_content(child_cst_node, result.content)

    return result


@_register(_TSNamedField.EMBEDDING_CONTENT)
def process_node_embedding_content(
        node: TSNode) -> CSTDocumentNodeContentEmbedding:
    lines = []
    for child in node.children:
        if child.type == _EMPTY_LINE_GRAMMAR_TYPE:
            lines.append('')
        elif child.text is not None:
            lines.append(child.text.decode())
        else:
            lines.append('')

    return CSTDocumentNodeContentEmbedding(
        content='\n'.join(lines))


@_register(_TSNamedField.LIST_OL)
def process_list_ordered(node: TSNode) -> CSTList:
    items: list[CSTListItem] = []
    for child_cst_node in _process_named_children(node):
        if isinstance(child_cst_node, CSTListItem):
            items.append(child_cst_node)
        else:
            raise TypeError('Invalid child of ordered list!', child_cst_node)

    return CSTList(type_=ListType.ORDERED, content=items)


@_register(_TSNamedField.LIST_UNOL)
def process_list_unordered(node: TSNode) -> CSTList:
    items: list[CSTListItem] = []
    for child_cst_node in _process_named_children(node):
        if isinstance(child_cst_node, CSTListItem):
            items.append(child_cst_node)
        else:
            raise TypeError('Invalid child of unordered list!', child_cst_node)

    return CSTList(type_=ListType.UNORDERED, content=items)


@_register(_TSNamedField.LIST_ITEM)
def process_list_item(node: TSNode) -> CSTListItem:
    index: int | None = None
    content: list[CSTEmptyLine | CSTList | CSTAnnotation | CSTRichtext] = []

    for child_cst_node in _process_named_children(node):
        if isinstance(
            child_cst_node,
            CSTEmptyLine | CSTList | CSTAnnotation | CSTRichtext
        ):
            _add_or_merge_content(child_cst_node, content)

        elif isinstance(child_cst_node, CSTListItemOLIndexContainer):
            if index is not None:
                raise TypeError(
                    'Multiple indices for list item', child_cst_node)

            index = child_cst_node.value
        else:
            raise TypeError('Invalid child of list item!', child_cst_node)

    return CSTListItem(index=index, content=content)


@_register(_TSNamedField.LIST_OL_INDEX)
def process_list_ol_index(node: TSNode) -> CSTListItemOLIndexContainer:
    if node.text is None:
        raise ValueError('Empty list index!', node)
    else:
        return CSTListItemOLIndexContainer(value=int(node.text.decode()))


@_register(_TSNamedField.EMPTY_LINE)
def process_empty_line(node: TSNode) -> CSTEmptyLine:
    if node.text is None:
        return CSTEmptyLine(content='')
    else:
        return CSTEmptyLine(content=node.text.decode())


@_register(_TSNamedField.NODE)
def process_document_node(node: TSNode) -> CSTDocumentNode:
    doc_node = CSTDocumentNode(
        title=[],
        metadata=[],
        content=None)

    for child_cst_node in _process_named_children(node):
        if isinstance(child_cst_node, CSTEmptyLine):
            if doc_node.metadata:
                doc_node.metadata.append(child_cst_node)
            else:
                doc_node.title.append(child_cst_node)

        elif isinstance(child_cst_node, CSTDocumentNodeMetadata):
            doc_node.metadata.extend(child_cst_node.content)

        elif isinstance(child_cst_node, CSTDocumentNodeTitle):
            doc_node.title.extend(child_cst_node.content)

        elif isinstance(
            child_cst_node,
            CSTDocumentNodeContentRichtext | CSTDocumentNodeContentEmbedding
        ):
            if doc_node.content is not None:
                raise ValueError('Document node with multiple contents!')
            doc_node.content = child_cst_node

        else:
            raise TypeError('Invalid child of document node!', child_cst_node)

    return doc_node


@_register(_TSNamedField.RICHTEXT_LINE)
def process_richtext_line(node: TSNode) -> CSTRichtext:
    """Note: this also handles everything with formatting autoclosing.
    """
    result = CSTRichtext(content=[])
    was_autoclose = (node.type == _AUTOCLOSE_GRAMMAR_TYPE)

    for child in node.children:
        if child.type == _TSRichtextTypes.PLAINTEXT:
            if child.text is None:
                result.content.append('')
            else:
                result.content.append(child.text.decode())

        elif child.type == _TSRichtextTypes.FMT_PRE:
            result.content.append(CSTFormattingMarker(
                marker=InlineFormatting.PRE,
                was_autoclose=was_autoclose))
        elif child.type == _TSRichtextTypes.FMT_UNDERLINE:
            result.content.append(CSTFormattingMarker(
                marker=InlineFormatting.UNDERLINE,
                was_autoclose=was_autoclose))
        elif child.type == _TSRichtextTypes.FMT_STRONG:
            result.content.append(CSTFormattingMarker(
                marker=InlineFormatting.STRONG,
                was_autoclose=was_autoclose))
        elif child.type == _TSRichtextTypes.FMT_EMPHASIS:
            result.content.append(CSTFormattingMarker(
                marker=InlineFormatting.EMPHASIS,
                was_autoclose=was_autoclose))
        elif child.type == _TSRichtextTypes.FMT_STRIKE:
            result.content.append(CSTFormattingMarker(
                marker=InlineFormatting.STRIKE,
                was_autoclose=was_autoclose))

        elif child.type in (
            _TSRichtextTypes.FMT_BRACKET_LINK_ANON,
            _TSRichtextTypes.FMT_BRACKET_LINK,
            _TSRichtextTypes.FMT_BRACKET_METADATA
        ):
            result.content.append(process_fmt_bracket(child))
        else:
            print(child.type)
            raise TypeError(
                'Invalid richtext line child type!',
                child, child.type, child.grammar_name)

    return result


@_register(_TSNamedField.RICHTEXT_FMT_BRACKET)
def process_fmt_bracket(  # noqa: C901, PLR0912, PLR0915
        node: TSNode
        ) -> CSTFmtBracketMetadata | CSTFmtBracketLink:
    if node.type == _TSRichtextTypes.FMT_BRACKET_LINK_ANON:
        target, *invalid_extras = _process_named_children(node)
        if invalid_extras:
            raise TypeError('Fmt bracket has invalid extras!', invalid_extras)
        if not isinstance(target, CSTMetadataTarget.__value__):
            raise TypeError('Fmt bracket has invalid target!', target)
        target = cast(CSTMetadataTarget, target)

        return CSTFmtBracketLink(
            content=None,
            target=target)

    elif node.type == _TSRichtextTypes.FMT_BRACKET_LINK:
        content: list[CSTRichtext] = []
        invalid_extras = []
        target = None
        for child in _process_named_children(node):
            if isinstance(child, CSTMetadataTarget.__value__):
                if target is not None:
                    raise TypeError(
                        'Fmt bracket has multiple targets!', child)
                # Note: only needed because of the weird isinstance hack above
                target = cast(CSTMetadataTarget, child)

            elif isinstance(child, CSTRichtext):
                _add_or_merge_content(child, content)

            else:
                invalid_extras.append(child)

        if invalid_extras:
            raise TypeError('Fmt bracket has invalid extras!', invalid_extras)
        if target is None:
            raise TypeError('Fmt bracket is missing target!', node)
        if len(content) > 1:
            raise TypeError('Fmt bracket has multiple richtexts!', content)

        if content:
            normalized_content, = content
        else:
            normalized_content = CSTRichtext(content=[])

        return CSTFmtBracketLink(
            content=normalized_content,
            target=target)

    elif node.type == _TSRichtextTypes.FMT_BRACKET_METADATA:
        content: list[CSTRichtext] = []
        invalid_extras = []
        metadata = None
        for child in _process_named_children(node):
            if isinstance(child, CSTFmtBracketMetadataContainer):
                if metadata is not None:
                    raise TypeError(
                        'Fmt bracket has multiple metadatas!', child)
                metadata = child

            elif isinstance(child, CSTRichtext):
                _add_or_merge_content(child, content)

            else:
                invalid_extras.append(child)

        if invalid_extras:
            raise TypeError('Fmt bracket has invalid extras!', invalid_extras)
        if metadata is None:
            raise TypeError('Fmt bracket is missing metadata!', node)
        if len(content) > 1:
            raise TypeError('Fmt bracket has multiple richtexts!', content)

        if content:
            normalized_content, = content
        else:
            normalized_content = CSTRichtext(content=[])

        return CSTFmtBracketMetadata(
            content=normalized_content,
            metadata=metadata.content)

    else:
        raise TypeError('Invalid fmt bracket node type!', node)


@_register(_TSNamedField.METADATA_ASSIGNMENT)
def process_metadata_assignment(node: TSNode) -> CSTMetadataAssignment:
    key, value, *invalid_extras = _process_named_children(node)
    if invalid_extras:
        raise TypeError(
            'Invalid child of metadata assignment!', invalid_extras)

    if not isinstance(key, CSTMetadataStr):
        raise TypeError('Metadata key is not metadata string!', key)
    if not isinstance(value, CSTMetadataValue.__value__):
        raise TypeError('Metadata value is not a metadata value type!', value)
    value = cast(CSTMetadataValue, value)

    return CSTMetadataAssignment(key=key, value=value)


@_register(_TSNamedField.METADATA_KEY)
def process_metadata_key(node: TSNode) -> str | CSTMetadataStr:
    return _normalize_sugarable_string(node)


@_register(_TSNamedField.METADATA_VALUE)
def process_metadata_value(  # noqa: C901, PLR0911, PLR0912
        node: TSNode
        ) -> CSTMetadataValue:
    first_child, *other_children = node.children

    if first_child.text is None:
        raise ValueError('Metadata values must have content!')

    all_child_types = {child.type for child in node.children}
    if len(all_child_types) != 1:
        raise TypeError('Metadata values must have exactly one type of value!')

    if first_child.type in (
        _TSValueTypes.STRING1, _TSValueTypes.STRING2
    ):
        value_type = cast(
            Literal[_TSValueTypes.STRING1] | Literal[_TSValueTypes.STRING2],
            _TSValueTypes(first_child.type))
        # Used to detect multi-line strings... which currently aren't working,
        # but shhhhh
        string_collector = []
        string_collector.append(first_child.text.decode())
        for other_child in other_children:
            if other_child.text is None:
                raise ValueError(
                    'Metadata values must have content!', other_child)

            string_collector.append(other_child.text.decode())

        return CSTMetadataStr(
            '\n'.join(string_collector),
            quote_style=_get_string_quote_style(value_type))

    elif other_children:
        raise ValueError(
            'Only multiline strings can have multiple metadata values!',
            other_children)

    elif first_child.type == _TSValueTypes.NULL:
        return CSTMetadataNull(value=None)
    elif first_child.type == _TSValueTypes.MISSING:
        return CSTMetadataMissing(value=None)
    elif first_child.type == _TSValueTypes.TRUE:
        return CSTMetadataBool(value=True)
    elif first_child.type == _TSValueTypes.FALSE:
        return CSTMetadataBool(value=False)
    elif first_child.type == _TSValueTypes.INT:
        return CSTMetadataInt(value=int(first_child.text.decode()))
    elif first_child.type == _TSValueTypes.DECIMAL:
        return CSTMetadataDecimal(value=Decimal(first_child.text.decode()))
    elif first_child.type == _TSValueTypes.MENTION:
        return CSTMetadataMention(
            value=_normalize_sugarable_string(first_child))
    elif first_child.type == _TSValueTypes.TAG:
        return CSTMetadataTag(
            value=_normalize_sugarable_string(first_child))
    elif first_child.type == _TSValueTypes.VARIABLE:
        return CSTMetadataVariable(
            value=_normalize_sugarable_string(first_child))
    elif first_child.type == _TSValueTypes.REF:
        return CSTMetadataReference(
            value=_normalize_sugarable_string(first_child))
    else:
        raise TypeError('Invalid metadata value type!', first_child)


@_register(_TSNamedField.RICHTEXT_FMT_TARGET)
def process_fmt_bracket_link_target(node: TSNode) -> CSTMetadataTarget:
    if node.type == _TSValueTypes.MENTION:
        return CSTMetadataMention(
            value=_normalize_sugarable_string(node))
    elif node.type == _TSValueTypes.TAG:
        return CSTMetadataTag(
            value=_normalize_sugarable_string(node))
    elif node.type == _TSValueTypes.VARIABLE:
        return CSTMetadataVariable(
            value=_normalize_sugarable_string(node))
    elif node.type == _TSValueTypes.REF:
        return CSTMetadataReference(
            value=_normalize_sugarable_string(node))
    elif node.type == _TSValueTypes.URI:
        return _normalize_sugarable_string(node)
    else:
        raise TypeError('Invalid format bracket link target!', node)


@_register(_TSNamedField.NODE_TITLE)
def process_node_title_container(node: TSNode) -> CSTDocumentNodeTitle:
    container = CSTDocumentNodeTitle(content=[])
    for child_cst_node in _process_named_children(node):
        if isinstance(child_cst_node, CSTRichtext | CSTEmptyLine):
            container.add_line(child_cst_node)
        else:
            raise TypeError(
                'Invalid child of node title container!', child_cst_node)

    return container


@_register(_TSNamedField.NODE_METADATA)
def process_node_metadata_container(node: TSNode) -> CSTDocumentNodeMetadata:
    container = CSTDocumentNodeMetadata(content=[])
    for child_cst_node in _process_named_children(node):
        if isinstance(
            child_cst_node,
            CSTMetadataAssignment | CSTAnnotation | CSTEmptyLine
        ):
            container.add_line(child_cst_node)
        else:
            raise TypeError(
                'Invalid child of node metadata container!', child_cst_node)

    return container


@_register(_TSNamedField.INLINE_METADATA)
def process_inline_metadata_container(
        node: TSNode) -> CSTFmtBracketMetadataContainer:
    container = CSTFmtBracketMetadataContainer(content=[])
    for child_cst_node in _process_named_children(node):
        if isinstance(child_cst_node, CSTMetadataAssignment):
            container.content.append(child_cst_node)
        else:
            raise TypeError(
                'Invalid child of inline metadata container!', child_cst_node)

    return container


@_register(_TSNamedField.ANNOTATION_LINE)
def process_annotation_line(node: TSNode) -> CSTAnnotation:
    content = []
    for child_name, child in _extract_named_children(node):
        if (
            child_name == _TSNamedField.ANNOTATION_LINE_CONTENT
            and child.text is not None
        ):
            content.append(child.text.decode())
    return CSTAnnotation(content=[''.join(content)])


def _process_named_children(ts_node: TSNode) -> Iterator[CSTNode]:
    for child_name, child in _extract_named_children(ts_node):
        maybe_cst_node = handle_node(child_name, child)
        if maybe_cst_node is not None:
            yield maybe_cst_node


def _extract_named_children(
        ts_node: TSNode) -> Iterator[tuple[_TSNamedField, TSNode]]:
    """This simply extracts the named children from the given TSNode,
    yielding them back in the same order as they appear in
    ts_node.children.
    """
    for index, child in enumerate(ts_node.children):
        field_name = ts_node.field_name_for_child(index)
        if field_name is not None:
            try:
                yield _TSNamedField(field_name), child
            except ValueError as exc:
                exc.add_note('Unknown field() annotation in grammar.js')
                exc.add_note(f'Grammar name: {ts_node.grammar_name}')
                exc.add_note(f'Grammar type: {ts_node.type}')
                exc.add_note(f'Child index: {index}')
                raise


def _add_or_merge_content(
        line: CSTNode | CSTRichtext | CSTAnnotation,
        # I'd like to be more specific than just plain list here, but mutable
        # collections are invariant, so that would defeat the purpose of
        # having a general-purpose method for merging things.
        # See: https://mypy.readthedocs.io/en/stable/generics.html#variance-of-generics
        content: list
        ) -> None:
    if content:
        previous_line = content[-1]
        if (
            isinstance(line, CSTRichtext)
            and isinstance(previous_line, CSTRichtext)
        ):
            previous_line.content.append(CSTLineBreak(content=None))
            previous_line.content.extend(line.content)
        # Note: the elif here (instead of an OR above) is purely so that the
        # type checker succeeds
        elif (
            isinstance(line, CSTAnnotation)
            and isinstance(previous_line, CSTAnnotation)
        ):
            previous_line.content.append(CSTLineBreak(content=None))
            previous_line.content.extend(line.content)

        else:
            content.append(line)

    else:
        content.append(line)


def _normalize_sugarable_string(node: TSNode) -> CSTMetadataStr:
    """Sugarable strings can be either a sugared string, which is a
    node with no children and only text, or a nested STRING1|STRING2
    within the children of the node.

    This normalizes and flattens both of them into the types expected by
    the metadata values.
    """
    if node.children:
        should_be_string = process_metadata_value(node)

        if not isinstance(should_be_string, CSTMetadataStr):
            raise TypeError(
                'Sugarable string was not a string?', should_be_string)
        return should_be_string
    elif node.text is None:
        raise ValueError('Sugarable string must have content!', node)
    else:
        return CSTMetadataStr(
            value=node.text.decode(),
            quote_style=None)



"""After getting the result of a parse from treesitter, we convert
everything into using instances of these objects instead. This gives us
stronger typing, a more consistent API, and a convenient place to put
any needed helper methods, in a way that doesn't get exposed as part
of the package API.
"""
from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from decimal import Decimal
from typing import Literal

from cleancopy.spectypes import InlineFormatting
from cleancopy.spectypes import ListType

# Note: URIs are automatically converted to strings; they're only separate in
# the CST because sugared strings are a strict subset of strings and need to
# be differentiated from the other target types in the grammar itself
type CSTMetadataTarget = (
    CSTMetadataStr | CSTMetadataMention | CSTMetadataTag | CSTMetadataVariable
    | CSTMetadataReference)
type CSTMetadataValue = (
    CSTMetadataStr | CSTMetadataInt | CSTMetadataDecimal | CSTMetadataBool
    | CSTMetadataNull | CSTMetadataMention | CSTMetadataTag
    | CSTMetadataVariable | CSTMetadataReference | CSTMetadataMissing)


@dataclass(kw_only=True)
class CSTNode:
    """Base class for all CST node types.
    TODO: put line and column infos here!
    """


@dataclass(kw_only=True)
class CSTDocument(CSTNode):
    root: CSTDocumentNode[CSTDocumentNodeContentRichtext]


@dataclass(kw_only=True)
class CSTDocumentNode[
        C: CSTDocumentNodeContentRichtext | CSTDocumentNodeContentEmbedding
        ](CSTNode):
    """The base class for both richtext and embedded nodes."""
    title: list[CSTRichtext | CSTEmptyLine]
    metadata: list[CSTMetadataAssignment | CSTAnnotation | CSTEmptyLine]
    content: C | None


@dataclass(kw_only=True)
class CSTDocumentNodeTitle(CSTNode):
    """This class is used purely for disambiguation when dealing with
    the output of treesitter. Its content gets elided into the parent
    CSTDocumentNode title attribute.
    """
    content: list[CSTRichtext | CSTEmptyLine]

    def add_line(
            self,
            line: CSTRichtext | CSTEmptyLine
            ) -> None:
        """This is a convenience method that handles merging multiple
        consecutive CSTRichtext instances into single values with
        line breaks between them.
        """
        if isinstance(line, CSTRichtext) and self.content:
            previous_line = self.content[-1]
            if isinstance(previous_line, CSTRichtext):
                previous_line.content.append(CSTLineBreak(content=None))
                previous_line.content.extend(line.content)
            else:
                self.content.append(line)

        else:
            self.content.append(line)


@dataclass(kw_only=True)
class CSTDocumentNodeMetadata(CSTNode):
    """This class is used purely for disambiguation when dealing with
    the output of treesitter. Its content gets elided into the parent
    CSTDocumentNode metadata attribute.
    """
    content: list[CSTMetadataAssignment | CSTAnnotation | CSTEmptyLine]

    def add_line(
            self,
            line: CSTMetadataAssignment | CSTAnnotation | CSTEmptyLine
            ) -> None:
        """This is a convenience method that handles merging multiple
        consecutive CSTAnnotation instances into single values with
        line breaks between them.
        """
        if isinstance(line, CSTAnnotation) and self.content:
            previous_line = self.content[-1]
            if isinstance(previous_line, CSTAnnotation):
                previous_line.content.append(CSTLineBreak(content=None))
                previous_line.content.extend(line.content)
            else:
                self.content.append(line)

        else:
            self.content.append(line)


@dataclass(kw_only=True)
class CSTDocumentNodeContentRichtext(CSTNode):
    content: list[
        CSTEmptyLine | CSTList | CSTAnnotation | CSTRichtext | CSTDocumentNode]


@dataclass(kw_only=True)
class CSTDocumentNodeContentEmbedding(CSTNode):
    content: str


@dataclass(kw_only=True)
class CSTEmptyLine(CSTNode):
    content: str


@dataclass(kw_only=True)
class CSTLineBreak(CSTNode):
    content: None


@dataclass(kw_only=True)
class CSTList(CSTNode):
    type_: ListType
    content: list[CSTListItem]


@dataclass(kw_only=True)
class CSTListItemOLIndexContainer(CSTNode):
    """This class is used purely for disambiguation when dealing with
    the output of treesitter. Its content gets elided into the parent
    CSTListItem index attribute.
    """
    value: int


@dataclass(kw_only=True)
class CSTListItem(CSTNode):
    index: int | None
    content: list[CSTEmptyLine | CSTList | CSTAnnotation | CSTRichtext]


@dataclass(kw_only=True)
class CSTRichtext(CSTNode):
    content: list[
        str | CSTFormattingMarker | CSTFmtBracketLink | CSTFmtBracketMetadata
        | CSTLineBreak]


@dataclass(kw_only=True)
class CSTFormattingMarker(CSTNode):
    marker: InlineFormatting
    was_autoclose: bool = field(default=False, kw_only=True)


@dataclass(kw_only=True)
class CSTAnnotation(CSTNode):
    content: list[str | CSTLineBreak]


@dataclass(kw_only=True)
class CSTFmtBracketMetadataContainer(CSTNode):
    """This class is used purely for disambiguation when dealing with
    the output of treesitter. Its content gets elided into the parent
    CSTInlineMetadata
    """
    content: list[CSTMetadataAssignment]


@dataclass(kw_only=True)
class CSTFmtBracketMetadata(CSTNode):
    content: CSTRichtext
    metadata: list[CSTMetadataAssignment]


@dataclass(kw_only=True)
class CSTFmtBracketLink(CSTNode):
    content: None | CSTRichtext
    target: CSTMetadataTarget


@dataclass
class CSTMetadataAssignment(CSTNode):
    key: CSTMetadataStr
    value: CSTMetadataValue


@dataclass
class CSTMetadataStr(CSTNode):
    value: str
    # Note: gets set to None for sugared strings
    quote_style: None | Literal[1] | Literal[2] = field(kw_only=True)


@dataclass
class CSTMetadataInt(CSTNode):
    value: int


@dataclass
class CSTMetadataDecimal(CSTNode):
    value: Decimal


@dataclass
class CSTMetadataBool(CSTNode):
    value: bool


@dataclass
class CSTMetadataNull(CSTNode):
    value: None


@dataclass
class CSTMetadataMissing(CSTNode):
    value: None


@dataclass
class CSTMetadataMention(CSTNode):
    value: CSTMetadataStr


@dataclass
class CSTMetadataTag(CSTNode):
    value: CSTMetadataStr


@dataclass
class CSTMetadataVariable(CSTNode):
    value: CSTMetadataStr


@dataclass
class CSTMetadataReference(CSTNode):
    value: CSTMetadataStr



"""This contains special-purpose adapters to convert from the CST to
the AST.
"""
from __future__ import annotations

import logging
from collections.abc import Iterator
from collections.abc import Sequence
from dataclasses import dataclass
from dataclasses import field
from functools import singledispatch
from typing import cast
from typing import overload

from cleancopy._cst.nodes import CSTAnnotation
from cleancopy._cst.nodes import CSTDocument
from cleancopy._cst.nodes import CSTDocumentNode
from cleancopy._cst.nodes import CSTDocumentNodeContentEmbedding
from cleancopy._cst.nodes import CSTDocumentNodeContentRichtext
from cleancopy._cst.nodes import CSTEmptyLine
from cleancopy._cst.nodes import CSTFmtBracketLink
from cleancopy._cst.nodes import CSTFmtBracketMetadata
from cleancopy._cst.nodes import CSTFormattingMarker
from cleancopy._cst.nodes import CSTLineBreak
from cleancopy._cst.nodes import CSTList
from cleancopy._cst.nodes import CSTListItem
from cleancopy._cst.nodes import CSTMetadataAssignment
from cleancopy._cst.nodes import CSTMetadataBool
from cleancopy._cst.nodes import CSTMetadataDecimal
from cleancopy._cst.nodes import CSTMetadataInt
from cleancopy._cst.nodes import CSTMetadataMention
from cleancopy._cst.nodes import CSTMetadataMissing
from cleancopy._cst.nodes import CSTMetadataNull
from cleancopy._cst.nodes import CSTMetadataReference
from cleancopy._cst.nodes import CSTMetadataStr
from cleancopy._cst.nodes import CSTMetadataTag
from cleancopy._cst.nodes import CSTMetadataVariable
from cleancopy._cst.nodes import CSTRichtext
from cleancopy.ast_.nodes import Annotation
from cleancopy.ast_.nodes import ASTNode
from cleancopy.ast_.nodes import BlockMetadata
from cleancopy.ast_.nodes import DocNode
from cleancopy.ast_.nodes import Document
from cleancopy.ast_.nodes import EmbeddingDocNode
from cleancopy.ast_.nodes import InlineFormatting
from cleancopy.ast_.nodes import InlineMetadata
from cleancopy.ast_.nodes import List_
from cleancopy.ast_.nodes import ListItem
from cleancopy.ast_.nodes import MetadataAssignment
from cleancopy.ast_.nodes import MetadataBool
from cleancopy.ast_.nodes import MetadataDecimal
from cleancopy.ast_.nodes import MetadataInt
from cleancopy.ast_.nodes import MetadataMention
from cleancopy.ast_.nodes import MetadataNull
from cleancopy.ast_.nodes import MetadataReference
from cleancopy.ast_.nodes import MetadataStr
from cleancopy.ast_.nodes import MetadataTag
from cleancopy.ast_.nodes import MetadataTarget
from cleancopy.ast_.nodes import MetadataValue
from cleancopy.ast_.nodes import MetadataVariable
from cleancopy.ast_.nodes import Paragraph
from cleancopy.ast_.nodes import Richtext
from cleancopy.ast_.nodes import RichtextDocNode
from cleancopy.exceptions import MultipleDocumentMetadatas
from cleancopy.spectypes import MetadataMagics

logger = logging.getLogger(__name__)


@singledispatch
def convert(cst_node, **kwargs) -> ASTNode | str | None:
    raise TypeError('No converter available for node!', cst_node)


@convert.register
def _(cst_node: CSTDocument, **kwargs) -> Document:
    cst_root_container = cst_node.root.content

    if cst_root_container is None:
        return Document(
            title=None, metadata=None, root=RichtextDocNode(content=[]))

    root_content = []
    root_node = RichtextDocNode(content=root_content)
    document = Document(title=None, metadata=None, root=root_node)
    for child_cst_node in cst_root_container.content:
        child_ast_node = convert(child_cst_node, **kwargs)

        if (
            isinstance(child_ast_node, RichtextDocNode)
            and child_ast_node.metadata is not None
            and child_ast_node.metadata.is_doc_metadata
        ):
            if document.metadata is not None:
                raise MultipleDocumentMetadatas()
            else:
                document.metadata = child_ast_node.metadata
                document.title = child_ast_node.title

        else:
            root_content.append(child_ast_node)

    return document


@convert.register
def _(
        cst_node: CSTDocumentNode, **kwargs
        ) -> RichtextDocNode | EmbeddingDocNode:
    # Awkwardly, we can have empty lines interleaved with richtexts
    merged_title_richtext_content = []
    for cst_title_node in cst_node.title:
        if isinstance(cst_title_node, CSTEmptyLine):
            # Let the dedicated function handle conversion instead of repeating
            # it here. The important thing is that, in a title line, we're
            # ignoring empty lines
            merged_title_richtext_content.append(
                convert(CSTLineBreak(content=None), **kwargs))
        else:
            merged_title_richtext_content.extend(cst_title_node.content)

    # A little weird, but we can just create a new virtual richtext item and
    # use it instead
    title = cast(
        Richtext,
        convert(CSTRichtext(content=merged_title_richtext_content), **kwargs))

    metadata = BlockMetadata()
    for cst_metadata_node in cst_node.metadata:
        if not isinstance(cst_metadata_node, CSTEmptyLine):
            ast_metadata_node = cast(
                MetadataAssignment | Annotation,
                convert(cst_metadata_node, **kwargs))
            metadata._add(ast_metadata_node)

    cst_node_content = cst_node.content
    if cst_node_content is None:
        if metadata.embed is None:
            return RichtextDocNode(title=title, metadata=metadata, content=[])
        else:
            return EmbeddingDocNode(
                title=title, metadata=metadata, content=None)

    elif isinstance(cst_node_content, CSTDocumentNodeContentRichtext):
        result = RichtextDocNode(
            title=title,
            metadata=metadata,
            content=[])
        convert(cst_node_content, into_node=result, **kwargs)
        return result

    elif isinstance(cst_node_content, CSTDocumentNodeContentEmbedding):
        return EmbeddingDocNode(
            title=title,
            metadata=metadata,
            content=cst_node_content.content)

    else:
        raise TypeError('Impossible branch: unknown CST doc node type!')


@convert.register
def _(
        cst_node: CSTDocumentNodeContentRichtext,
        *, into_node: RichtextDocNode, **kwargs) -> RichtextDocNode:
    """Given a partially-processed RichtextNode with an empty content,
    converts the CST's content into AST form and appends it to the
    into_node, and then returns the node back.
    """
    for cst_nodegroup_or_nested_node in _group_by_paragraph(cst_node.content):
        if isinstance(cst_nodegroup_or_nested_node, CSTDocumentNode):
            into_node.content.append(
                cast(
                    DocNode,
                    convert(cst_nodegroup_or_nested_node, **kwargs)))

        # Status check: now we have
        # list[CSTList | CSTAnnotation | CSTRichtext], corresponding to a
        # single paragraph. Note that CSTAnnotations and CSTRichtexts have
        # already consolidated multiple lines; you don't need to do that again!
        else:
            this_paragraph = Paragraph(content=[])

            for child_cst_node in cst_nodegroup_or_nested_node:
                child_cst_node = cast(
                    CSTList | CSTAnnotation | CSTRichtext, child_cst_node)
                this_paragraph.content.append(
                    cast(
                        Richtext | List_ | Annotation,
                        convert(child_cst_node, **kwargs)))

            into_node.content.append(this_paragraph)

    return into_node


@dataclass
class _FmtStackState:
    current_span: Richtext
    fmt_marker: InlineFormatting | None
    to_join: list[str] = field(default_factory=list)


@convert.register
def _(cst_node: CSTRichtext, **kwargs) -> Richtext:
    outermost_span = Richtext(
        context=None,
        content=[])
    fmt_stack: list[_FmtStackState] = [_FmtStackState(outermost_span, None)]

    for child_cst_node in cst_node.content:
        if isinstance(child_cst_node, str):
            fmt_stack[-1].to_join.append(child_cst_node)
        elif isinstance(child_cst_node, CSTLineBreak):
            fmt_stack[-1].to_join.append(
                cast(str, convert(child_cst_node, **kwargs)))

        # Status check:
        # child_cst_node:
        #   CSTFormattingMarker | CSTFmtBracketLink | CSTFmtBracketMetadata
        else:
            stale_stack_state = fmt_stack[-1]

            if stale_stack_state.to_join:
                span_text = ''.join(stale_stack_state.to_join)
                stale_stack_state.to_join.clear()
                stale_stack_state.current_span.content.append(span_text)

            if isinstance(child_cst_node, CSTFormattingMarker):
                this_fmt_marker = child_cst_node.marker
                if this_fmt_marker == stale_stack_state.fmt_marker:
                    span_to_pop = fmt_stack.pop()
                    if span_to_pop.to_join:
                        span_to_pop.current_span.content.append(
                            ''.join(span_to_pop.to_join))
                else:
                    nested_richtext = Richtext(
                        context=InlineMetadata(),
                        content=[])
                    nested_richtext.context._add(MetadataAssignment(
                        key=MetadataMagics.sugared.value,
                        value=MetadataBool(value=True)))
                    nested_richtext.context._add(MetadataAssignment(
                        key=MetadataMagics.fmt.value,
                        value=MetadataStr(value=this_fmt_marker)))
                    fmt_stack.append(_FmtStackState(
                        current_span=nested_richtext,
                        fmt_marker=this_fmt_marker))
                    stale_stack_state.current_span.content.append(
                        nested_richtext)

            else:
                # Note: this doesn't get added to the formatting stack, because
                # it's a whole self-contained thing
                nested_richtext = cast(
                    Richtext, convert(child_cst_node, **kwargs))
                stale_stack_state.current_span.content.append(nested_richtext)
                # fmt_stack.append(_FmtStackState(
                #     current_span=nested_richtext,
                #     fmt_marker=None))

    if len(fmt_stack) != 1 or fmt_stack[-1].current_span is not outermost_span:
        print(fmt_stack)
        raise RuntimeError('Failed to fully exhaust richtext fmt stack!')
    outermost_stack_state, = fmt_stack
    outermost_span.content.append(''.join(outermost_stack_state.to_join))

    # Note: because we've had BOTH the stack AND been constructing the Richtext
    # tree as we go along, we don't need to manipulate the tree at all; we
    # already have it fully constructed and can simply return the outermost
    # span.
    return outermost_span


@convert.register
def _(cst_node: CSTList, **kwargs) -> List_:
    result = List_(type_=cst_node.type_, content=[])

    for cst_list_item in cst_node.content:
        result.content.append(cast(ListItem, convert(cst_list_item, **kwargs)))

    return result


@convert.register
def _(cst_node: CSTListItem, **kwargs) -> ListItem:
    result = ListItem(index=cst_node.index, content=[])

    # Reminder: we've got an iterator of
    # list[CSTList | CSTAnnotation | CSTRichtext], corresponding to a
    # single paragraph. Note that CSTAnnotations and CSTRichtexts have
    # already consolidated multiple lines; you don't need to do that again!
    for cst_nodegroup in _group_by_paragraph(cst_node.content):
        this_paragraph = Paragraph(content=[])

        for child_cst_node in cst_nodegroup:
            child_cst_node = cast(
                CSTList | CSTAnnotation | CSTRichtext, child_cst_node)
            this_paragraph.content.append(
                cast(
                    Richtext | List_ | Annotation,
                    convert(child_cst_node, **kwargs)))

        result.content.append(this_paragraph)

    return result


@convert.register
def _(cst_node: CSTFmtBracketMetadata, **kwargs) -> Richtext:
    context = InlineMetadata()
    for child_metadata in cst_node.metadata:
        context._add(
            cast(MetadataAssignment, convert(child_metadata, **kwargs)))

    span_content = cst_node.content
    if span_content is None:
        return Richtext(context=context, content=[])
    else:
        return Richtext(
            context=context,
            content=[cast(Richtext, convert(span_content, **kwargs))])


@convert.register
def _(cst_node: CSTFmtBracketLink, **kwargs) -> Richtext:
    target = cast(MetadataTarget, convert(cst_node.target, **kwargs))
    context = InlineMetadata()
    context._add(MetadataAssignment(
        key=MetadataMagics.sugared.value, value=MetadataBool(value=True)))
    context._add(MetadataAssignment(
        key=MetadataMagics.target.value, value=target))

    span_content = cst_node.content
    if span_content is None:
        return Richtext(context=context, content=[])
    else:
        return Richtext(
            context=context,
            content=[cast(Richtext, convert(span_content, **kwargs))])


@convert.register
def _(cst_node: CSTMetadataAssignment, **kwargs) -> MetadataAssignment:
    return MetadataAssignment(
        key=cst_node.key.value,
        value=cast(MetadataValue, convert(cst_node.value)))


@convert.register
def _(cst_node: CSTMetadataStr, **kwargs) -> MetadataStr:
    return MetadataStr(value=cst_node.value)


@convert.register
def _(cst_node: CSTMetadataInt, **kwargs) -> MetadataInt:
    return MetadataInt(value=cst_node.value)


@convert.register
def _(cst_node: CSTMetadataDecimal, **kwargs) -> MetadataDecimal:
    return MetadataDecimal(value=cst_node.value)


@convert.register
def _(cst_node: CSTMetadataBool, **kwargs) -> MetadataBool:
    return MetadataBool(value=cst_node.value)


@convert.register
def _(cst_node: CSTMetadataNull, **kwargs) -> MetadataNull:
    return MetadataNull(value=cst_node.value)


@convert.register
def _(cst_node: CSTMetadataMissing, **kwargs) -> None:
    return None


@convert.register
def _(cst_node: CSTMetadataMention, **kwargs) -> MetadataMention:
    return MetadataMention(value=cst_node.value.value)


@convert.register
def _(cst_node: CSTMetadataTag, **kwargs) -> MetadataTag:
    return MetadataTag(value=cst_node.value.value)


@convert.register
def _(cst_node: CSTMetadataVariable, **kwargs) -> MetadataVariable:
    return MetadataVariable(value=cst_node.value.value)


@convert.register
def _(cst_node: CSTMetadataReference, **kwargs) -> MetadataReference:
    return MetadataReference(value=cst_node.value.value)


@convert.register
def _(cst_node: CSTAnnotation, **kwargs) -> Annotation:
    to_join: list[str] = []
    for maybe_line_break in cst_node.content:
        if isinstance(maybe_line_break, CSTLineBreak):
            stringified = cast(str, convert(maybe_line_break, **kwargs))
            to_join.append(stringified)
        else:
            to_join.append(maybe_line_break)

    return Annotation(content=''.join(to_join))


@convert.register
def _(
        cst_node: CSTLineBreak, *,
        replace_linebreak_with: str = ' ',
        **kwargs) -> str:
    # Note: collapsing multiple line breaks is part of the actual grammar,
    # so we don't -- or at least shouldn't -- need to worry about it here
    return replace_linebreak_with


@overload
def _group_by_paragraph[
            T: CSTEmptyLine | CSTList | CSTAnnotation | CSTRichtext](
        lines: Sequence[T]
        ) -> Iterator[list[T]]: ...
@overload
def _group_by_paragraph[
            T: CSTEmptyLine | CSTList | CSTAnnotation | CSTRichtext](
        lines: Sequence[T | CSTDocumentNode]
        ) -> Iterator[list[T] | CSTDocumentNode]: ...
def _group_by_paragraph[
            T: CSTEmptyLine | CSTList | CSTAnnotation | CSTRichtext](
        lines: Sequence[T | CSTDocumentNode]
        ) -> Iterator[list[T] | CSTDocumentNode]:
    this_paragraph = []
    for line in lines:
        if isinstance(line, CSTEmptyLine):
            if this_paragraph:
                yield this_paragraph
                this_paragraph = []

        elif isinstance(line, CSTDocumentNode):
            if this_paragraph:
                yield this_paragraph
                this_paragraph = []
            yield line

        else:
            this_paragraph.append(line)

    yield this_paragraph



from __future__ import annotations

import logging
import re
from dataclasses import dataclass
from dataclasses import fields
from decimal import Decimal
from typing import Any
from typing import ClassVar
from typing import Optional

from cleancopy.spectypes import EmbedFallbackBehavior
from cleancopy.spectypes import InlineFormatting
from cleancopy.spectypes import ListType
from cleancopy.spectypes import MetadataMagics

# Note: URIs are automatically converted to strings; they're only separate in
# the CST because sugared strings are a strict subset of strings and need to
# be differentiated from the other target types in the grammar itself
type MetadataTarget = (
    MetadataStr | MetadataMention | MetadataTag | MetadataVariable
    | MetadataReference)
# This is used to omit reserved but unused metadata keys from the results. We
# do this so people don't accidentally try to use them, and then later on
# we have to worry about compatibility issues if we need to introduce new
# fields.
METADATA_MAGIC_PATTERN = re.compile(r'^__.+__$')
logger = logging.getLogger(__name__)


@dataclass(kw_only=True)
class ASTNode:
    """Currently not really used for anything except for annotations,
    but at any rate: this is the base class for all AST nodes.
    """


@dataclass(kw_only=True)
class Document(ASTNode):
    # Note: this comes from the __doc_meta__ node
    title: Richtext | None
    metadata: BlockMetadata | None
    root: RichtextDocNode
    # TODO: add other helper methods, like searching by ID


@dataclass(kw_only=True)
class DocNode(ASTNode):
    """The base class for both richtext and embedded nodes."""
    title: Optional[Richtext] = None
    metadata: Optional[BlockMetadata] = None


@dataclass(kw_only=True)
class RichtextDocNode(DocNode):
    content: list[Paragraph | DocNode]

    def __getitem__(self, index: int) -> Paragraph | DocNode:
        return self.content[index]

    def __len__(self):
        return len(self.content)

    def __iter__(self):
        return iter(self.content)


@dataclass(kw_only=True)
class EmbeddingDocNode(DocNode):
    # Can be an explicit None if it's an empty node
    content: str | None


@dataclass(kw_only=True)
class Paragraph(ASTNode):
    """Paragraphs can contain multiple lines and/or multiple line types,
    but they **cannot** contain an empty line.
    """
    # Note: these are separate because lists have their own separate
    # RichtextContext for items
    content: list[Richtext | List_ | Annotation]

    def __bool__(self) -> bool:
        """Returns True if the paragraph has any content at all, whether
        displayed or not displayed.
        """
        return bool(self.content)


@dataclass(kw_only=True)
class List_(ASTNode):  # noqa: N801
    type_: ListType
    content: list[ListItem]


@dataclass(kw_only=True)
class ListItem(ASTNode):
    index: int | None
    content: list[Paragraph]


@dataclass(kw_only=True)
class Richtext[C: InlineMetadata | None](ASTNode):
    context: C
    content: list[str | Richtext]

    @property
    def has_display_content(self) -> bool:
        """Returns True if the paragraph contains any non-annotation
        lines. If all lines are annotations, or there are no lines,
        returns False.
        """
        if not self.content:
            return False
        else:
            return not all(
                isinstance(segment, Annotation) for segment in self.content)


@dataclass(kw_only=True)
class Annotation(ASTNode):
    content: str


# Note: can't be protocol due to missing intersection type
class _MemoizedFieldNames:
    _field_names: ClassVar[frozenset[str]]

    @staticmethod
    def memoize[C: type](for_cls: C) -> C:
        for_cls._field_names = frozenset(
            {field.name for field in fields(for_cls)})
        return for_cls


@_MemoizedFieldNames.memoize
@dataclass(kw_only=True)
class Metadata(ASTNode, _MemoizedFieldNames):

    # TODO: uri vs ref vs mention vs...
    target: Optional[MetadataTarget] = None

    def __post_init__(self):
        self._lookup: dict[str, MetadataValue | None] = {}
        self._payload: list

    # TODO: we should add items, .get, etc
    def __getitem__(self, key: str) -> MetadataValue:
        """Note: THIS is where we normalize things re: missings, None,
        etc.
        """
        maybe_result = self._lookup[key]

        if maybe_result is None:
            raise KeyError(
                'Metadata key was explicitly set to __missing__', key)

        else:
            return maybe_result

    def _add(self, line: MetadataAssignment | Annotation):
        """Call this when building the AST. Intended for use within the
        CST -> AST transition. So... semi-public. Public in the sense
        that it's used outside of this module, but not public in the
        sense that it's documented or intended for outside use.
        """
        self._payload.append(line)
        if isinstance(line, MetadataAssignment):
            key = line.key

            try:
                metadata_magic = MetadataMagics(key)

            except ValueError:
                if METADATA_MAGIC_PATTERN.match(key) is None:
                    self._lookup[key] = line.value
                else:
                    logger.warning('Ignoring reserved metadata key: %s', key)

            else:
                maybe_field_name = metadata_magic.name
                if maybe_field_name in self._field_names:
                    maybe_value = line.value
                    # Note that we want to extract the actual metadata value;
                    # the magics are all strongly-typed, so we don't need to
                    # worry about the MetadataValue container around them.
                    # (even though it's preserved within .as_declared)
                    if maybe_value is None:
                        value_to_use = None
                    else:
                        value_to_use = maybe_value.value

                    setattr(self, maybe_field_name, value_to_use)
                else:
                    logger.warning(
                        'Wrong metadata type for reserved key %s; ignoring',
                        key)


@_MemoizedFieldNames.memoize
@dataclass(kw_only=True)
class InlineMetadata(Metadata):
    """InlineMetadata is used only for, yknow, inline metadata.
    Note that all of the various formatting tags get sugared into
    inline metadatas.
    """
    icu_1: Optional[MetadataReference] = None
    fmt: Optional[InlineFormatting] = None
    sugared: Optional[bool] = None

    def __post_init__(self):
        super().__post_init__()
        self._payload: list[MetadataAssignment] = []

    @property
    def as_declared(self) -> tuple[MetadataAssignment, ...]:
        """This can be used to access the raw assignments, as declared,
        in their exact order. For inline metadata, this is only relevant
        if there are multiple metadata assignments using the same key
        in the same InlineMetadata instance.
        """
        # The only reason we do a tuple here is to make sure that the outside
        # world doesn't try to modify this!
        return tuple(self._payload)


@_MemoizedFieldNames.memoize
@dataclass(kw_only=True)
class BlockMetadata(Metadata):
    """BlockMetadata is used both for node metadata and for document
    metadata (which is itself just an empty node at the toplevel with
    a special magic key set).
    """
    is_doc_metadata: bool = False
    id_: Optional[MetadataStr] = None
    embed: Optional[MetadataStr] = None
    fallback: Optional[EmbedFallbackBehavior] = None

    def __post_init__(self):
        super().__post_init__()
        self._payload: list[MetadataAssignment | Annotation] = []

    @property
    def as_declared(self) -> tuple[MetadataAssignment | Annotation, ...]:
        """This can be used to access the raw assignments, as declared,
        in their exact order, **including any annotations**. This is
        useful both if you want access to the annotations, ^^or^^ if
        there are multiple metadata assignments using the same key in
        the same BlockMetadata instance.
        """
        # The only reason we do a tuple here is to make sure that the outside
        # world doesn't try to modify this!
        return tuple(self._payload)


@dataclass
class MetadataAssignment(ASTNode):
    key: str
    value: MetadataValue | None


@dataclass
class MetadataValue(ASTNode):
    # Note: needs to be overridden by subclasses
    value: Any


@dataclass
class MetadataStr(MetadataValue):
    value: str


@dataclass
class MetadataInt(MetadataValue):
    value: int


@dataclass
class MetadataDecimal(MetadataValue):
    value: Decimal


@dataclass
class MetadataBool(MetadataValue):
    value: bool


@dataclass
class MetadataNull(MetadataValue):
    value: None


@dataclass
class MetadataMention(MetadataValue):
    value: str


@dataclass
class MetadataTag(MetadataValue):
    value: str


@dataclass
class MetadataVariable(MetadataValue):
    value: str


@dataclass
class MetadataReference(MetadataValue):
    value: str



"""This module contains additional types, constants, and literals
associated with the language spec itself.
"""
from __future__ import annotations

from enum import Enum
from enum import StrEnum


class InlineFormatting(StrEnum):
    PRE = '__pre__'
    UNDERLINE = '__underline__'
    STRONG = '__strong__'
    EMPHASIS = '__emphasis__'
    STRIKE = '__strike__'


# These string values are not part of the spec -- they're purely an
# implementation detail -- hence Enum
class ListType(Enum):
    ORDERED = 'ORDERED'
    UNORDERED = 'UNORDERED'


# These string values are literals that are part of the spec, hence StrEnum
class EmbedFallbackBehavior(StrEnum):
    VISIBLE_PREFORMATTED = 'VISIBLE_PREFORMATTED'
    HIDDEN = 'HIDDEN'


class MetadataMagics(Enum):
    # Note: the field names here need to match the field names in ast.nodes
    id_ = '__id__'
    embed = '__embed__'
    target = '__target__'
    icu_1 = '__icu-1__'
    fmt = '__fmt__'
    sugared = '__sugared__'
    is_doc_metadata = '__doc_meta__'
    fallback = '__fallback__'



class CleancopyException(Exception):
    """This, or a subclass of this, is raised for all expected internal
    exceptions that are used for control flow in the library.
    """


class InvalidCleancopy(Exception):
    """This, or a subclass of this, is raise for all expected external
    exceptions -- ie, any problems with the document itself.
    """


class MultipleDocumentMetadatas(InvalidCleancopy):
    """Raised when we encounter multiple nodes marked as document
    metadata.
    """
